sub template_save() {
	my ($filename,$template)=@_;
	my (@point,$y,$x);

	if(open(OUT,"$filename")) {
		foreach $one(@{$template}) {
			print OUT "[template]\n";
			print OUT "name=$one->{name}\n";
			print OUT "width=$one->{width}\n";
			print OUT "height=$one->{height}\n";
			print OUT "base_x=$one->{base_x}\n" if(exists $one->{base_x});
			print OUT "base_y=$one->{base_y}\n" if(exists $one->{base_y});
			print OUT "center_x=$one->{center_x}\n" if(exists $one->{center_x});
			print OUT "center_y=$one->{center_y}\n" if(exists $one->{center_y});
			print OUT "encount=$one->{encount}\n" if(exists $one->{encount});
			print OUT "num_pixel=$one->{num_pixel}\n";
			print OUT "point={";
			@point=();
			foreach $y(0..$one->{height}-1) {
				foreach $x(0..$one->{width}-1) {
					if($one->{point}[$y][$x]==1) {
						push @point,"($x,$y)";
					}
				}
			}
			print OUT join(",",@point);
			print OUT "}\n\n";
		}
		close OUT;
	}
}

sub template_load() {
	my ($filename)=@_;
	my ($key,$value);
	my ($width,$height,$number);
	my (@template,@point,$xy,$x,$y);

	if(open(IN,"<$filename")) {
		$number=-1;
		foreach(<IN>) {
			s/[\r\n]//g;
			if($_ eq '[template]') {
				$number++;
			}
			elsif(/^#/) {
				# 
			}
			elsif(/^([^=]+)=(.*)/) {
				$key=$1;
				$value=$2;
				if($key eq 'name') {
					$template[$number]->{name}=$value;
				}
				elsif($key eq 'width') {
					$template[$number]->{width}=$value;
					$width=$value;
				}
				elsif($key eq 'height') {
					$template[$number]->{height}=$value;
					$height=$value;
				}
				elsif($key eq 'num_pixel') {
					$template[$number]->{num_pixel}=$value;
				}
				elsif($key eq 'point') {
					$value=~s/\{\(//;
					$value=~s/\)\}//;
					@point=split(/\),\(/,$value);

					# ٥󥰺ɸ
					foreach $y(0..$height-1) {
						foreach $x(0..$width-1) {
							$template[$number]->{point}[$y][$x]=0;
						}
					}

					# ٥󥰺ɸ
					foreach $xy(@point) {
						($x,$y)=split(/,/,$xy);
						$template[$number]->{point}[$y][$x]=1;
					}
				}
			}
		}
		close IN;

		return [@template];
	}
}

sub template_to_ppm() {
	my ($template,$red,$green,$blue)=@_;
	my ($ppm,$width,$height,$maxpixel,$x,$y);

	$width=$template->{width};
	$height=$template->{height};
	$maxpixel=255;

	$ppm->{width}=$width;
	$ppm->{height}=$height;
	$ppm->{maxpixel}=$maxpixel;

	foreach $y(0..$height-1) {
		foreach $x(0..$width-1) {
			if($template->{point}[$y][$x]==1) {
				$ppm->{pixel}[$y][$x]->{R}=$red;
				$ppm->{pixel}[$y][$x]->{G}=$green;
				$ppm->{pixel}[$y][$x]->{B}=$blue;
			}
			else {
				$ppm->{pixel}[$y][$x]->{R}=0;
				$ppm->{pixel}[$y][$x]->{G}=0;
				$ppm->{pixel}[$y][$x]->{B}=0;
			}
		}
	}

	return $ppm;
}

sub template_merge() {
	my ($template_t)=@_;
	my (%unique,$one,@point,$y,$x);
	my (@new_t);

	%unique=();
	foreach $one(@{$template_t}) {
		@point=();
		foreach $y(0..$one->{height}-1) {
			foreach $x(0..$one->{width}-1) {
				if($one->{point}[$y][$x]==1) {
					push @point,"($x,$y)";
				}
			}
		}
		if(exists $unique{join(',',@point)}) {
			$unique{join(',',@point)}->{encount}++;
		}
		else {
			push @new_t,$one;
			$unique{join(',',@point)}=$one;
			$unique{join(',',@point)}->{encount}=1;
		}
	}
	@new_t=sort{$a->{name} cmp $b->{name}}@new_t;
	return [@new_t];
}

sub template_dynamic_match() {
	my ($ref_image_t,$ref_template_t,$threshold)=@_;
	my (@img,@t,$image_t,$template_t);
	my ($ci,$ct,$xyi,$xyt);
	my ($expected_name,@score,$match,$maximum,$name);

	foreach $image_t(@{$ref_image_t}) {
		@img=@{$image_t->{point}};
		@score=();
		foreach $template_t(@{$ref_template_t}) {
			@t=@{$template_t->{point}};
			$name=$template_t->{name};

			# ʥߥåޥå
			$gap=1;
			$wi=$image_t->{width};
			$hi=$image_t->{height};
			$wt=$template_t->{width};
			$ht=$template_t->{height};
			$ci=$image_t->{num_pixel};
			$ct=$template_t->{num_pixel};

			#print "img=$image_t->{name} ($ci:$wi,$hi), t=$name ($ct:$wt,$ht)\n";

			$window_x=&min($wi,$wt);
			$window_y=&min($hi,$ht);
			for $wy(-$gap+0..+$gap+abs($hi-$ht)) {
				for $wx(-$gap+0..+$gap+abs($wi-$wt)) {
					#print "b=($wx,$wy)\n";
					$match=0;
					@point=();
					for $by(0..$window_y-1) {
						for $bx(0..$window_x-1) {
							#print "  w=($bx,$by)\n";
							if($wi>$wt) {
								$img_x=$wx+$bx;
								$t_x=$bx;
								next if($img_x<0 || $img_x>=$wi);
							}
							else {
								$img_x=$bx;
								$t_x=$wx+$bx;
								next if($t_x<0 || $t_x>=$wt);
							}
							if($hi>$ht) {
								$img_y=$wy+$by;
								$t_y=$by;
								next if($img_y<0 || $img_y>=$hi);
							}
							else {
								$img_y=$by;
								$t_y=$wy+$by;
								next if($t_y<0 || $t_y>=$ht);
							}
							#print "    img($img_x,$img_y)=$img[$img_y][$img_x] && t($t_x,$t_y)=$t[$t_y][$t_x]\n";
							if($img[$img_y][$img_x]==1 && $t[$t_y][$t_x]==1) {
								push @point,{img_x=>$img_x,img_y=>$img_y,t_x=>$t_x,t_y=>$t_y};
								$match++;
							}
						}
					}
					if($match/$ct>=$threshold) {
						$cx=&median(map{$_->{img_x}}@point);
						$cy=&median(map{$_->{img_y}}@point);
						push @score,{name=>$name,score=>$match/$ct,center_x=>$cx,center_y=>$cy};
						#print "  img[$image_t->{name}] ($ci:$wi,$hi), t[$name] ($ct:$wt,$ht)\n";
						#print "  one score: $match/$ct=",$match/$ct,", center=($cx,$cy)\n";
					}
				}
			}
		}
		@score=sort{$a->{center_x}<=>$b->{center_x}}@score;
		$image_t->{name}=join('',map{$_->{name}}@score);
		#print "  expected: \"$image_t->{name}\"\n";
		#print "  score=",join(',',map{$_->{score}}@score),"\n";
		$image_t->{score}=&mean(map{$_->{score}}@score);
		#print "  mean=$image_t->{score}\n";

		# ­ǤͤǤʤ
		last if($image_t->{score}<$threshold);
	}

	$expected_name=join('',map{$_->{name}}@{$ref_image_t});

	# ͽ¬줿̾ʸǤʤ
	#die if(length($expected_name)!=4);

	return $expected_name;
}

sub template_dynamic_match_retry() {
	my ($ref_image_t,$ref_template_t)=@_;
	my ($len,$expected_name,$threshold);
	my (%flag);

	$threshold=0.97;
	while(1) {
		$expected_name=&template_dynamic_match($ref_image_t,$ref_template_t,$threshold);
		$len=length $expected_name;
		if($len==4) {
			last;
		}
		elsif($len<4) {
			$threshold-=0.005;
			print "  retry <<< down <<< $threshold\n";
			$flag{down}=undef;
		}
		else {
			$threshold+=0.005;
			print "  retry >>> up >>>  $threshold\n";
			$flag{up}=undef;
		}
		# ʸͽ¬ʤ
		return if(exists $flag{up} && exists $flag{down});
	}
	return $expected_name;
}

1;
